from copy import deepcopy
from jinja2.exceptions import TemplateError
from requests.exceptions import RequestException


METHODS = ('GET', 'OPTIONS', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE',)
EARLIER_ERRORS_SKIP = 'Request skipped due to earlier error.'
NO_HTTP_METHOD = (
    'Request definition should contain exactly one HTTP method as '
    'a main level dict key.')


class RequestState:
    SUCCESS = 'SUCCESS'
    NOT_RAISED = 'NOT-RAISED'
    FAILURE = 'FAILURE'
    ERROR = 'ERROR'
    SKIPPED = 'SKIPPED'

    def __init__(self, state, message=None):
        states = (
            self.SUCCESS,
            self.NOT_RAISED,
            self.FAILURE,
            self.ERROR,
            self.SKIPPED)

        if state not in states:
            raise ValueError(
                'Unknown state.'
                f'State must be {", ".join(states[:-1])} or {states[-1]}.')

        self.state = state
        self.message = message

    def __eq__(self, other):
        if isinstance(other, RequestState):
            return self.state == other.state

        return self.state == other

    def __str__(self):
        return self.state

    @property
    def ok(self):
        return self.state in (self.SUCCESS, self.NOT_RAISED, self.SKIPPED)


class RequestOptions:
    def __init__(self, request_dict):
        self.register = request_dict.get('register')
        self.raise_for_status = request_dict.get('raise_for_status', True)
        self.output = request_dict.get('output')


class Assertion:
    def __init__(self, raw_assertion):
        self._ok = None
        if isinstance(raw_assertion, dict):
            self.name = raw_assertion.get('name')
            self.expression = raw_assertion.get('expression')
        else:
            self.name = raw_assertion
            self.expression = raw_assertion

    @property
    def executed(self):
        return self._ok is not None

    @property
    def ok(self):
        if self._ok is None:
            raise RuntimeError('Assertion has not been executed yet.')
        return self._ok

    def execute(self, template_env):
        try:
            self._ok = bool(template_env.resolve_expression(self.expression))
        except BaseException:
            self._ok = False
            raise

        return self._ok


class Request:
    def __init__(self, request_dict, template_env, skip=False):
        self._raw = deepcopy(request_dict)
        self._processed = None
        self._template_env = template_env

        self.state = None
        self.response = None

        self._parse_assertions()

        if skip:
            self._set_state(RequestState.SKIPPED, EARLIER_ERRORS_SKIP)
        else:
            self._process_templates()

        self.name = self._request.pop('name', None)
        self._parse_method_and_params()
        self.options = RequestOptions(self._request)

    @property
    def _request(self):
        return self._processed or self._raw

    def _set_state(self, state, message=None):
        self.state = RequestState(state, message)

    def _process_templates(self):
        try:
            self._processed = self._template_env.resolve_templates(
                self._request)
        except TemplateError as error:
            self._set_state(RequestState.ERROR, message=str(error))

    def _parse_method_and_params(self):
        method_keys = [
            key for key in self._request.keys() if key.upper() in METHODS]
        if not method_keys or len(method_keys) > 1:
            self._set_state(
                RequestState.ERROR,
                NO_HTTP_METHOD)
            return

        self.method = method_keys[0].upper()
        self.params = self._request.pop(method_keys[0])

    def _parse_assertions(self):
        raw_assertions = self._request.pop('assert', [])
        if not isinstance(raw_assertions, list):
            raw_assertions = [raw_assertions]

        self.assertions = [
            Assertion(raw_assertion) for raw_assertion in raw_assertions]

    def send(self, request_function):
        if self.state is not None:
            return

        try:
            self.response = request_function(self.method, **self.params)
        except RequestException as error:
            self._set_state(RequestState.ERROR, str(error))
            return

        self._template_env.register('response', self.response)
        if self.options.register:
            self._template_env.register(self.options.register, self.response)

        if not self.response.ok:
            if self.options.raise_for_status:
                self._set_state(RequestState.FAILURE)
            else:
                self._set_state(RequestState.NOT_RAISED)
        else:
            self._set_state(RequestState.SUCCESS)

        for assertion in self.assertions:
            try:
                ok = assertion.execute(self._template_env)
                if not ok:
                    self._set_state(RequestState.FAILURE)
            except BaseException as error:
                self._set_state(RequestState.ERROR, message=str(error))



from os import system
import platform
from traceback import print_exc

from jinja2 import __version__ as _jinja2_version
from yaml import __version__ as _pyyaml_version
from requests import __version__ as _requests_version

from .utils.args import get_argparser, load_plan_file, parse_variables
from ._logger import RequestLogger
from ._plan import Plan
from ._runner import PlanRunner
from ._version import __version__


UNKNOWN_ERROR_MSG = (
    'Caught unexpected error, see traceback and description below. '
    'If this seems like a bug to you, please consider creating a issue in '
    'https://github.com/kangasta/yaml_requests/issues.\n'
)


def _print_versions():
    print(
        f'{__version__} ('
        f'jinja2={_jinja2_version}',
        f'pyyaml={_pyyaml_version}',
        f'requests={_requests_version}'
        ')')


NO_PLAN = 251
INVALID_PLAN = 252
INTERRUPTED = 253
UNKNOWN_ERROR = 254


def main():
    # On windows, run shell command to enable ANSI code support
    if platform.system() == 'Windows':
        system('')

    args = get_argparser().parse_args()
    logger = RequestLogger(animations=args.animation, colors=args.colors)

    try:
        if args.version:
            _print_versions()
            return 0

        if not args.plan_file:
            logger.error('No requests plan file provided.')
            return NO_PLAN

        variables_override = parse_variables(args.variables)

        try:
            plan_dict = load_plan_file(args.plan_file)
            plan = Plan(plan_dict, variables_override=variables_override)
        except FileNotFoundError:
            logger.error(f'Did not find plan file in {args.plan_file}.')
            return NO_PLAN
        except (ValueError, AssertionError,) as error:
            logger.error(str(error))
            return INVALID_PLAN

        runner = PlanRunner(plan, logger)
        num_errors = runner.run()
        return min(num_errors, 250)
    except KeyboardInterrupt:
        logger.stop_progress_animation()
        return INTERRUPTED
    except BaseException:
        logger.stop_progress_animation()
        logger.error(UNKNOWN_ERROR_MSG)
        print_exc()
        return UNKNOWN_ERROR



import json
import yaml
import re
from shutil import get_terminal_size
from threading import Event, Thread

from ._request import RequestState

# From cli-spinners (https://www.npmjs.com/package/cli-spinners)
INTERVAL = 0.080  # seconds
FRAMES = ['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏']

COLORS = dict(red=31, green=32, yellow=33, blue=34, grey=90)


def _fit_to_width(text):
    max_len = get_terminal_size().columns
    non_formatted_text = re.sub(r'\033\[[0-9]+m', '', text.replace('\r', ''))
    if len(non_formatted_text) < max_len:
        return text

    i = 0
    j = 0
    while i < max_len - 2:
        match = re.match(r'\r|\033\[[0-9]+m', text[(i + j):])
        if match:
            j += len(match.group(0))
        else:
            i += 1

    clear_formatting = '\033[0m' if re.search(r'\033\[[0-9]+m', text) else ''
    return f'{text[:(i + j)]}{clear_formatting}…'


def _print_spinner_and_text(text, stop_event):
    i = 0
    while not stop_event.wait(INTERVAL):
        print(_fit_to_width(f'\r{FRAMES[i % len(FRAMES)]} {text}'), end='')
        i += 1
    print('\r', end='')


def get_assertion_status(assertion):
    if not assertion.executed:
        return RequestState.SKIPPED
    elif assertion.ok:
        return RequestState.SUCCESS
    else:
        return RequestState.ERROR


def get_indicator(state):
    if state in (RequestState.SUCCESS, RequestState.NOT_RAISED,):
        return ('green', '✔',)
    elif state in (RequestState.FAILURE, RequestState.ERROR,):
        return ('red', '✘',)
    elif state == RequestState.SKIPPED:
        return ('blue', '⮟',)


class RequestLogger:
    def __init__(self, animations, colors):
        self._animations = animations
        self._colors = colors

        self._active = None
        self._stop_event = Event()

    def bold(self, text):
        if not self._colors:
            return text
        return f'\033[1m{text}\033[22m'

    def color(self, text, color):
        if not self._colors or color not in COLORS:
            return text
        return f'\033[{COLORS[color]}m{text}\033[0m'

    def error(self, text):
        error_text = self.bold(self.color('ERROR:', 'red'))
        print(f'{error_text} {text}')

    def title(self, name, num_requests):
        name_text = f'{self.bold(name)}\n' if name else ''
        print(f'{name_text}Sending {num_requests} requests:\n')

    def _get_indicator_text(self, request=None, assertion=None):
        color, symbol = get_indicator(
            request.state if request else get_assertion_status(assertion))
        return self.color(symbol, color)

    def _get_name_text(self, request):
        if request.name:
            return self.bold(request.name)
        return ''

    def _get_method_text(self, request):
        return f'{self.bold(request.method)} {request.params.get("url")}'

    def _get_response_code_text(self, request):
        response = request.response
        if response is None:
            return ''

        code_text = self.bold(f'HTTP {response.status_code}')
        elapsed_ms = response.elapsed.total_seconds() * 1000
        not_raised_text = ''
        if request.state == RequestState.NOT_RAISED:
            not_raised_text = self.color(
                ' HTTP status code ignored.', 'grey')

        return f'{code_text} ({elapsed_ms:.3f} ms){not_raised_text}'

    def _get_message_text(self, request):
        message = request.state.message

        type_text = self.bold(f'{str(request.state).upper()}:')
        return f'{type_text} {message}' if message else ''

    def _get_assertion_text(self, request):
        text = ''

        if request.state == RequestState.SKIPPED:
            return text

        for assertion in request.assertions:
            indicator = self._get_indicator_text(assertion=assertion)
            text += f'\n  {indicator} {assertion.name}'

        return f'{text}\n' if text else ''

    def _headers_text(self, response):
        return '\n'.join(
            f'{self.bold(key)}: {value}' for key,
            value in response.headers.items())

    def _response_output_text(self, response, output):
        def _format_output(output):
            if not output.endswith('\n'):
                output = f'{output}\n'
            output = output.replace('\n', '\n  ').rstrip(' ')
            return f'\n  {output}'

        try:
            if not output:
                return ''
            elif output.lower() == 'headers':
                return _format_output(self._headers_text(response))
            elif output.lower() == 'text':
                return _format_output(response.text)
            elif output.lower() == 'json':
                pretty_json = json.dumps(response.json(), indent=2)
                return _format_output(pretty_json)
            elif output.lower() in ('yml', 'yaml'):
                pretty_yaml = yaml.dump(response.json())
                return _format_output(pretty_yaml)
            else:
                return ''
        except BaseException:
            return ''

    def _response_text(self, request):
        output = request.options.output
        output = output if isinstance(output, list) else [output]
        response = request.response

        return ''.join(self._response_output_text(response, i) for i in output)

    def start_request(self, request):
        if not self._animations:
            return

        text = self._get_name_text(request) or self._get_method_text(request)

        self._active = Thread(
            target=_print_spinner_and_text, args=[
                text, self._stop_event])
        self._stop_event.clear()
        self._active.start()

    def stop_progress_animation(self):
        self._stop_event.set()
        if self._active:
            self._active.join()
            self._active = None

    def finish_request(self, request):
        self.stop_progress_animation()

        name_text = self._get_name_text(request)
        name_separator = name_text and '\n  '
        code_text = self._get_response_code_text(request)
        message_text = self._get_message_text(request)
        message_separator = '\n  ' if message_text and code_text else ''

        text = (
            f'{self._get_indicator_text(request)} {name_text}{name_separator}'
            f'{self._get_method_text(request)}\n  '
            f'{code_text}{message_separator}{message_text}\n'
            f'{self._get_assertion_text(request)}'
            f'{self._response_text(request)}')

        print(text)



from copy import deepcopy


class PlanOptions:
    def __init__(self, plan_dict, options_override=None):
        if options_override is None:
            options_override = {}

        self._options = {**plan_dict.get('options', {}), **options_override}

        self.session = self._options.get('session', False)
        self.ignore_errors = self._options.get('ignore_errors')


class Plan:
    def __init__(
            self,
            plan_dict,
            options_override=None,
            variables_override=None):
        if variables_override is None:
            variables_override = {}

        self._plan = deepcopy(plan_dict)

        self.name = self._plan.pop('name', None)
        self.options = PlanOptions(self._plan, options_override)
        self.variables = {
            **self._plan.get('variables', {}), **variables_override}
        self.requests = self._plan.get('requests')

        if not self.requests or not isinstance(self.requests, list):
            raise AssertionError('Plan must contain requests array.')



from requests import request, Session
from requests.cookies import cookiejar_from_dict

from .utils.template import Environment
from ._request import Request


class PlanRunner:
    def __init__(self, plan, logger):
        self._plan = plan
        self._env = Environment()
        self._prepare_session()

        for name, value in self._plan.variables.items():
            self._env.register(name, value)

        self._logger = logger

    def _prepare_session(self):
        session_option = self._plan.options.session
        if not session_option:
            self._session = None
            return

        self._session = Session()

        if isinstance(session_option, dict):
            session_dict = self._env.resolve_templates(session_option)
            self._session.headers.update(session_dict.get('headers', {}))
            self._session.cookies.update(
                cookiejar_from_dict(
                    session_dict.get(
                        'cookies', {})))

    def _request(self, *args, **kwargs):
        if self._plan.options.session:
            return self._session.request(*args, **kwargs)
        return request(*args, **kwargs)

    def run(self):
        self._logger.title(self._plan.name, len(self._plan.requests))

        num_errors = 0

        for request_dict in self._plan.requests:
            skip = not self._plan.options.ignore_errors and num_errors > 0

            request = Request(request_dict, self._env, skip)

            if request.state is None:
                self._logger.start_request(request)
                request.send(self._request)

            self._logger.finish_request(request)

            if not request.state.ok:
                num_errors += 1

        return num_errors
